//
// BindStatusCallback.h
//
// Transparent Control OCX
// Tom Armstrong & Mark Nelson
// September 1997
//
// The BindStatusCallback class is used to monitor the
// progress of an asynchronous download.  Unfortunately,
// getting the ATL version of this class to work properly
// with our app is very difficult.  The object does a
// good job of passing information to the control as the
// download proceeds, but there is no special action taken
// when the download is complete.
//
// The solution to this problem is this customized version
// of the class, called COurBindStatusCallback.  It differs
// in two key areas.  First, the callback function it
// uses to notify the control has an additional parameter that
// tells the control whether a transfer is initiating, in
// progress, or complete.  Second, it has an extra call to
// ensure that the control is notified properly when the
// download is complete.
//
//

template <class T>
class ATL_NO_VTABLE COurBindStatusCallback :
    public CComObjectRootEx<T::_ThreadModel::ThreadModelNoCS>,
    public IBindStatusCallbackImpl<T>
{
 // -twa- Add the BSCF enum to the prototype
  typedef void (T::*ATL_PDATAAVAILABLE)(COurBindStatusCallback<T>* pbsc,
                                        DWORD grfBSCF,
                                        BYTE* pBytes,
                                        DWORD dwSize );
public:

BEGIN_COM_MAP(COurBindStatusCallback<T>)
  COM_INTERFACE_ENTRY_IID(IID_IBindStatusCallback, IBindStatusCallbackImpl<T>)
END_COM_MAP()

  COurBindStatusCallback()
  {
    m_pT = NULL;
    m_pFunc = NULL;
  }
  ~COurBindStatusCallback()
  {
    ATLTRACE(_T("~COurBindStatusCallback\n"));
  }

  STDMETHOD(OnStartBinding)(DWORD dwReserved, IBinding *pBinding)
  {
    ATLTRACE(_T("COurBindStatusCallback::OnStartBinding\n"));
    m_spBinding = pBinding;
    return S_OK;
  }

  STDMETHOD(GetPriority)(LONG *pnPriority)
  {
    ATLTRACENOTIMPL(_T("COurBindStatusCallback::GetPriority"));
  }

  STDMETHOD(OnLowResource)(DWORD reserved)
  {
    ATLTRACENOTIMPL(_T("COurBindStatusCallback::OnLowResource"));
  }

  STDMETHOD(OnProgress)(ULONG ulProgress,
                        ULONG ulProgressMax,
                        ULONG ulStatusCode,
                        LPCWSTR szStatusText)
  {
    ATLTRACENOTIMPL(_T("COurBindStatusCallback::OnProgress"));
  }

  STDMETHOD(OnStopBinding)(HRESULT hresult, LPCWSTR szError)
  {
    ATLTRACE(_T("COurBindStatusCallback::OnStopBinding\n"));
    m_spBinding.Release();
    return S_OK;
  }

  STDMETHOD(GetBindInfo)(DWORD *pgrfBINDF, BINDINFO *pbindInfo)
  {
    ATLTRACE(_T("COurBindStatusCallback::GetBindInfo\n"));
    *pgrfBINDF = BINDF_ASYNCHRONOUS | BINDF_ASYNCSTORAGE |
                 BINDF_GETNEWESTVERSION | BINDF_NOWRITECACHE;
    pbindInfo->cbSize = sizeof(BINDINFO);
    pbindInfo->szExtraInfo = NULL;
    memset(&pbindInfo->stgmedData, 0, sizeof(STGMEDIUM));
    pbindInfo->grfBindInfoF = 0;
    pbindInfo->dwBindVerb = BINDVERB_GET;
    pbindInfo->szCustomVerb = NULL;
    return S_OK;
  }

  STDMETHOD(OnDataAvailable)(DWORD grfBSCF,
                             DWORD dwSize,
                             FORMATETC *pformatetc,
                             STGMEDIUM *pstgmed)
  {
    ATLTRACE(_T("COurBindStatusCallback::OnDataAvailable\n"));
    HRESULT hr = S_OK;
    //
    // Get the Stream passed
    //
    if ( BSCF_FIRSTDATANOTIFICATION & grfBSCF )
    {
      if (!m_spStream && pstgmed->tymed == TYMED_ISTREAM)
      {
        m_spStream = pstgmed->pstm;
        if (m_spStream)
          m_spStream->AddRef();
      }
    }

    DWORD dwRead = dwSize - m_dwTotalRead; // Minimum amount available that hasn't been read
    DWORD dwActuallyRead = 0;            // Placeholder for amount read during this pull

    // If there is some data to be read then go ahead and read them
    if (m_spStream)
    {
      if (dwRead > 0)
      {
        BYTE* pBytes = NULL;
        ATLTRY(pBytes = new BYTE[dwRead + 1]);
        if (pBytes == NULL)
          return S_FALSE;
        hr = m_spStream->Read(pBytes, dwRead, &dwActuallyRead);
        if (SUCCEEDED(hr))
        {
          pBytes[dwActuallyRead] = 0;
          if (dwActuallyRead>0)
          {
                  // -twa- Pass the BSCF enum to the control
            (m_pT->*m_pFunc)(this, grfBSCF, pBytes, dwActuallyRead);
            m_dwTotalRead += dwActuallyRead;
          }
        }
        delete[] pBytes;
      }
    }

    if (BSCF_LASTDATANOTIFICATION & grfBSCF)
    {
         // -twa- Make sure the control is notified
      if ( dwActuallyRead == 0 ) // -twa-
       (m_pT->*m_pFunc)(this, grfBSCF, 0, 0); // -twa-
      m_spStream.Release();
    }
    return hr;
  }

  STDMETHOD(OnObjectAvailable)(REFIID riid, IUnknown *punk)
  {
    ATLTRACENOTIMPL(_T("COurBindStatusCallback::OnObjectAvailable"));
  }

  HRESULT _StartAsyncDownload(BSTR bstrURL, IUnknown* pUnkContainer, BOOL bRelative)
  {
    m_dwTotalRead = 0;
    m_dwAvailableToRead = 0;
    HRESULT hr = S_OK;
    CComQIPtr<IServiceProvider, &IID_IServiceProvider> spServiceProvider(pUnkContainer);
    CComPtr<IBindHost> spBindHost;
    if (spServiceProvider)
      spServiceProvider->QueryService(SID_IBindHost, IID_IBindHost, (void**)&spBindHost);

    if (spBindHost == NULL)
    {
      if (bRelative)
        return E_NOINTERFACE;  // relative asked for, but no IBindHost
      hr = CreateURLMoniker(NULL, bstrURL, &m_spMoniker);
      if (SUCCEEDED(hr))
        hr = CreateBindCtx(0, &m_spBindCtx);

      if (SUCCEEDED(hr))
        hr = RegisterBindStatusCallback(m_spBindCtx, reinterpret_cast<IBindStatusCallback*>(static_cast<IBindStatusCallbackImpl<T>*>(this)), 0, 0L);
      else
        m_spMoniker.Release();

      if (SUCCEEDED(hr))
      {
        IStream* pStream;
        hr = m_spMoniker->BindToStorage(m_spBindCtx, 0, IID_IStream, (void**)&pStream);
      }
    }
    else
    {
      hr = CreateBindCtx(0, &m_spBindCtx);
      if (SUCCEEDED(hr))
        hr = RegisterBindStatusCallback(m_spBindCtx, reinterpret_cast<IBindStatusCallback*>(static_cast<IBindStatusCallbackImpl<T>*>(this)), 0, 0L);

      if (SUCCEEDED(hr))
      {
        if (bRelative)
          hr = spBindHost->CreateMoniker(bstrURL, m_spBindCtx, &m_spMoniker, 0);
        else
          hr = CreateURLMoniker(NULL, bstrURL, &m_spMoniker);
      }

      if (SUCCEEDED(hr))
      {
        IStream* pStream;
        hr = spBindHost->MonikerBindToStorage(m_spMoniker, NULL, reinterpret_cast<IBindStatusCallback*>(static_cast<IBindStatusCallbackImpl<T>*>(this)), IID_IStream, (void**)&pStream);
        ATLTRACE(_T("Bound"));
      }
    }
    return hr;
  }

 HRESULT StartAsyncDownload(T* pT, ATL_PDATAAVAILABLE pFunc, BSTR bstrURL, IUnknown* pUnkContainer = NULL, BOOL bRelative = FALSE)
 {
    m_pT = pT;
    m_pFunc = pFunc;
    return  _StartAsyncDownload(bstrURL, pUnkContainer, bRelative);
  }
  //
  // This is the important function that is actually called to get
  // download moving.
  //
  static HRESULT Download(T* pT, ATL_PDATAAVAILABLE pFunc, BSTR bstrURL, IUnknown* pUnkContainer = NULL, BOOL bRelative = FALSE)
  {
    CComObject<COurBindStatusCallback<T> > *pbsc;
    HRESULT hRes = CComObject<COurBindStatusCallback<T> >::CreateInstance(&pbsc);
    if (FAILED(hRes))
      return hRes;
    return pbsc->StartAsyncDownload(pT, pFunc, bstrURL, pUnkContainer, bRelative);
 }

  CComPtr<IMoniker> m_spMoniker;
  CComPtr<IBindCtx> m_spBindCtx;
  CComPtr<IBinding> m_spBinding;
  CComPtr<IStream> m_spStream;
  T* m_pT;
  ATL_PDATAAVAILABLE m_pFunc;
  DWORD m_dwTotalRead;
  DWORD m_dwAvailableToRead;
};
