/***************************************************************************
 *                                                                         *
 *       The information  contained herein is  PROPRIETARY  and            *
 *          CONFIDENTIAL to NewMonics Inc.  Patents pending.               *
 *                                                                         *
 *       Copyright (C) 1997 NewMonics Inc., All Rights reserved.           *
 *                                                                         *
 ***************************************************************************/

package demo.BlackJack;

import java.lang.String;

public class Hand
{
  //
  // the largest number of cards in a hand that sums to less than or
  // equal to 21 is 12: Ace, Ace, Ace, Ace, 2, 2, 2, 2, 3, 3, 3.
  //
  private static final int MAX_HAND_SIZE = 12;

  private Card[]  cards;
  private int     num_cards;
  private int     num_aces;
  private boolean hidden_card;

  //
  // for each combination of aces (0 up to the 4 existing ones)
  // there are only 2 possible ways to add them up and remain
  // under 21
  //
  private static final int[][] ace_values =
  {
    { 0, 0  },    // no aces
    { 1, 11 },    // 1 ace
    { 2, 12 },    // 2 aces
    { 3, 13 },    // 3 aces
    { 4, 14 }     // 4 aces
  };

  //
  // this can be a hand with a card face down
  //
  public Hand(boolean hidden)
  {
    cards = new Card[ MAX_HAND_SIZE ];
    hidden_card = hidden;
    empty();
  }

  //
  // this is used in the toString() method
  //
  // note that this code is not multithread safe
  //
  static char[] buf = new char[ MAX_HAND_SIZE * 3];

  //
  // a String representing the whole hand
  //
  public String toString()
  {
    int i = 0;
    int j = 0;

    if (num_cards == 0)
    {
      // Need to create a String here, so the "client" can delete it.
      return new String(buf, 0, 0);
    }

    if (hidden_card)
    {
      buf[i++] = '*';
      j++;
    }

    for ( ; j < num_cards; j++)
    {
      if (i != 0)
      {
        buf[i++] = ' ';
      }

      buf[i++] = cards[j].rank;
      buf[i++] = cards[j].suit;
    }

    return new String(buf, 0, i);
  }

  //
  // empty the hand
  //
  public void empty()
  {
    num_cards = 0;
    num_aces  = 0;
  }

  //
  // adds a card to the hand
  //
  public void addCard(Card c)
  {
    if (c.value == 1)
    {
      num_aces++;
    }

    cards[num_cards++] = c;
  } 

  //
  // calculate the value of a hand
  //
  public int calcValue()
  {
    if (num_cards == 0)
    {
      return 0;
    }
    else
    {
      return bestValue(cards, 0, num_cards - 1);
    }
  } 

  //
  // hide first card
  //
  public final void hide()
  {
    hidden_card = true;
  }

  //
  // reveal the first card
  //
  public final void reveal()
  {
    hidden_card = false;
  }

  //
  // we try to get as close to 21, without exceeding it
  // start and end are real indexes into the array
  //
  private int bestValue(Card[] cards, int start, int end)
  {
    //
    // first, get the value of everything but the aces
    //
    int   base = 0;
    for (int i = start; i <= end; ++i)
    {
      if (cards[i].value == 1)
      {
        continue;
      }
      base += cards[i].value;
    }

    //
    // the array is constructed such that the first element is
    // always <= the second element
    //
    int smaller = base + ace_values[num_aces][0];
    int larger  = base + ace_values[num_aces][1];

    if (larger <= 21)
    {
      return larger;
    }
    else
    {
      return smaller;
    }
  }

  //
  // display the hand of the graphic display
  //
  public void displayCards(int y_pos)
  {
    int xloc = 121;
    int yloc = y_pos;
    int inc  = 115;
	  
    if (num_cards == 0)
    {
      return;
    }

    // this means we are dealing with the dealer's hand
    if (hidden_card)
    {
      BlackJack.output.drawCard(cards[0], true, xloc, yloc);
      xloc += inc;
      buf[0] = '*';
    }

    for (int i = hidden_card ? 1 : 0; i < num_cards; i++)
    {
      BlackJack.output.drawCard(cards[i], false, xloc, yloc);
      xloc += inc;
    }
  }
}
