import java.awt.Color;
import java.awt.Component;
import java.awt.OrderableContainer;
import java.awt.Dimension;
import java.awt.AWTEvent;
import java.awt.AWTEventMulticaster;
import java.awt.event.ActionListener;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.Graphics;
import java.awt.Point;
import java.util.Enumeration;

/*
 *
 * PieChart
 * A lightweight component implementing a dynamically updatable pie chart.
 * Each "slice" has a string associated with it and can have its own mouse
 * events.
 *
 */
public class PieChart extends OrderableContainer implements ComponentListener
{

	int totalValue = 0;			// sum of all slice values
	boolean paintFlag = false;	// whether we WANT to update the slices or not.
	boolean dirty = false;		// whether we HAVE to update the slices or not.
	int minAngle = 3;			// minimum size, in degrees, of a slice
								// (otherwise goes into other)
	float startingAngle = 0f;	// starting angle.
	Slice other;				// "other" slice. always at location zero.
	MouseListener mlistener;	// default mouse listener(s) to add to all new slices.
	ActionListener alistener;	// default action listener(s) to add to all new slices.
	Component list;				// a component to repaint whenever the piechart is updated.
								// (most likely a list.)
	boolean writeLabels;		// whether or not newly-added slices should be displaying
								// their labels.

	//
	// Constructor
	//
	public PieChart()
	{ this("", null, null, true); }

	public PieChart(String label)
	{ this(label, null, null, true); }
	
	public PieChart(String label, boolean writeLabels)
	{ this(label, null, null, writeLabels); }

	public PieChart(ActionListener alistener, MouseListener mlistener)
	{ this("", alistener, mlistener, true); }

	public PieChart(String label, ActionListener alistener, MouseListener mlistener)
	{ this(label, alistener, mlistener, true); }

	public PieChart(String label, ActionListener alistener, MouseListener mlistener, boolean writeLabels)
	{
		setName(label);
		this.alistener = alistener;
		this.mlistener = mlistener;
		this.writeLabels = writeLabels;
		other=addSlice("other", 0, Color.black);
		addComponentListener(this);
	}

	//
	// Methods to manipulate slices.
	//
	public synchronized Slice addSlice(String name, int value, Color color)
	{
		return addSlice(new Slice(name, value, color));
	}
	public synchronized Slice addSlice(Slice s)
	{
		totalValue += s.getValue();
		s.setIndex(getComponentCount());
		add(s);
		dirty = dirty || (s.getValue() != 0);
		Dimension size = getSize();
		s.setSize(size.width, size.height);
		s.addMouseListener(mlistener);
		s.addActionListener(alistener);
		s.setDrawLabel(writeLabels);
		return s;
	}
	public synchronized Slice getSlice(String name)
	{
		for (int i=0; i<getComponentCount(); ++i) {
			Slice s = (Slice) getComponent(i);
			if (name.equals(s.getLabel())) return s;
		}
		return null;
	}
	public synchronized Slice getSlice(int i)
	{
		if ((i >= getComponentCount()) ||
			(i < 0)) return null;
		return (Slice)getComponent(i);
	}
	public synchronized Slice getSlice(Slice s)
	{
		for (int i=0; i<getComponentCount(); ++i) 
			if (s == getComponent(i)) return s;
		return null;
	}
	public void removeSlice(String name)	{ removeSlice(getSlice(name));	}
	// TODO: efficiently check that this slice is in our pieChart.
	public synchronized void removeSlice(Slice s)
	{
		if (s != null) {
			totalValue -= s.getValue();
			remove(s);
			dirty = dirty || (s.getValue() != 0);
		}
		// correct indices.
		for (int i=s.getIndex(); i<getComponentCount(); ++i) {
			Slice sl = (Slice)getComponent(i);
			sl.setIndex(i);
		}
	}
	public synchronized void removeAllSlices()
	{
		totalValue = 0;
		super.removeAll();
		other=addSlice("other", 0, Color.black);
		dirty=true;
	}
	public void changeSlice(String name, int value)	{ changeSlice(getSlice(name), value);	}
	// TODO: efficiently check that this slice is in our pieChart.
	public synchronized void changeSlice(Slice s, int value)
	{
		if (s != null) {
			totalValue = totalValue - s.getValue() + value;
			s.setValue(value);
			dirty = true;
		}
	}
	public void incSlice(String name)	{ incSlice(getSlice(name));	}
	// TODO: efficiently check that this slice is in our pieChart.
	public void incSlice(Slice s)
	{
		if (s != null) {
			s.incValue();
			++totalValue;
			dirty = true;
		}
	}

	//
	// Methods to add/remove listeners.
	//
	public synchronized void addActionListener(ActionListener listener)
	{
		alistener = AWTEventMulticaster.add(alistener, listener);
		for (int i=0; i<getComponentCount(); ++i) {
			Slice s = (Slice)getComponent(i);
			s.addActionListener(listener);
		}
	}
	public synchronized void removeActionListener(ActionListener listener)
	{
		alistener = AWTEventMulticaster.remove(alistener, listener);
		for (int i=0; i<getComponentCount(); ++i){
			Slice s = (Slice)getComponent(i);
			s.removeActionListener(listener);
		}
	}
	public synchronized void addMouseListener(MouseListener listener)
	{
		mlistener = AWTEventMulticaster.add(mlistener, listener);
		for (int i=0; i<getComponentCount(); ++i) getComponent(i).addMouseListener(listener);
	}
	public synchronized void removeMouseListener(MouseListener listener)
	{
		mlistener = AWTEventMulticaster.remove(mlistener, listener);
		for (int i=0; i<getComponentCount(); ++i) getComponent(i).removeMouseListener(listener);
	}

	//
	// Preferred / minimum Size methods
	//
	public Dimension getPreferredSize()
	{
		return new Dimension (250, 250);
	}
	public Dimension getMinimumSize() 
	{
		return new Dimension (25, 25);
	}

	//
	// Set/get minimum angles.
	//
	public int getMinAngle()	{ return minAngle; }
	public void setMinAngle(int minAngle)	{ this.minAngle = minAngle; dirty = true; }

	//
	// Set/get list component.
	//
	public Component getList()	{ return list; }
	public void setList(Component list)	{ this.list = list; }

	//
	// Method to update slice angles based on values and repaint.
	//
	public synchronized void updateSlices()
	{
		float startAngle = startingAngle, arcAngle;

		dirty = false;
		other.value = 0;

		//int total=0;
		//for (int i=0; i<slices.length; ++i) total+= ((Slice)(slices[i])).getValue();
		//if (total!=totalValue) System.out.println("totals are off! cached: "+totalValue+" actual: "+total); totalValue=total;
	
		for (int i=1; i<getComponentCount(); ++i) {
			Slice s = (Slice)getComponent(i);
			arcAngle = (360f * s.getValue() / totalValue);

			//if (s==other) System.out.println("other slice not in correct location!");
			//System.out.println("Slice "+(i+1)+": "+startAngle+","+arcAngle);
			if(arcAngle >= minAngle) {
				s.setAngles(startAngle, arcAngle+1);
				startAngle += arcAngle;
				s.setVisible(true);
			} else {
				other.value += s.getValue();
				s.setVisible(false);
			}
		}
		// the "other" slice gets the rest.
		int where = (int)startAngle;
		other.setAngles(where, 360+startingAngle-where);
		//repaint();
	}

	//
	// Method to check whether a point lies in the circle.
	//
	public boolean contains(Point p)	{ return contains(p.x, p.y);	}
	public boolean contains(int x, int y)
	{
		Dimension size = getSize();
		int cx = size.width / 2;
		int cy = size.height / 2;
		return (((x-cx)*(x-cx) + (y-cy)*(y-cy)) <= cx*cx);
	}


	//
	// When the size of the piechart changes, we want to change the sizes
	// of all of the slices.
	//
	public synchronized void componentResized(ComponentEvent e)
	{
		Dimension size = getSize();
		//System.out.println("piechart "+getName()+" resized to " + size.width + "x" + size.height);
		int newsize = Math.min(size.width, size.height);
		for (int i=0; i<getComponentCount(); ++i) {
			Component s = getComponent(i);
			s.setSize(newsize, newsize);
		}
	}

	// other events we have to implement because of ComponentListener.
	public void componentHidden(ComponentEvent e)
	{
		//System.out.println("piechart "+getName()+" hidden.");
	}
	public void componentMoved(ComponentEvent e)
	{
		//System.out.println("piechart "+getName()+" moved.");
	}
	public void componentShown(ComponentEvent e)
	{
		//System.out.println("piechart "+getName()+" shown.");
	}

	//
	// Swaps the order of two slices.
	//
	public synchronized void swap(int a, int b)
	{
		// swap components
		super.swap(a, b);
		// swap indices
		Slice slicea = (Slice)getComponent(b); Slice sliceb = (Slice)getComponent(a);
		int t=slicea.getIndex();
		slicea.setIndex(sliceb.getIndex());
		sliceb.setIndex(t);
		// make sure that no one is selected
		if (slicea.getSelected()) slicea.setSelected(false);
		if (sliceb.getSelected()) sliceb.setSelected(false);
	}

	public void disablePaint()	{ paintFlag = false; }
	public void enablePaint()	{ paintFlag = true; //repaint();
	}

	//
	// paint methods
	//
	public void paint(Graphics g)
	{
		//System.out.println("paint called: "+getName()+" paint: "+paintFlag+" dirty: "+dirty);
		if (paintFlag && dirty) { updateSlices(); if (list != null) list.repaint(); }
		super.paint(g);
	}
	public void update(Graphics g)
	{
		//System.out.println("update called: "+getName()+" paint: "+paintFlag+" dirty: "+dirty);
		paint(g);
	}
}
