/*
 * ProfileApplet.java
 *
 * John Whaley 11 Jul 1997
 */

import jx.EventMulticaster;
import jx.EventListener;
import java.applet.*;
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.net.*;
import java.awt.image.*;
import jx.profiler.*;

/*
 * WindowHandler
 *
 * Instances of this class are used to handle window openings/closings. It allows
 * windows to have a button associated with them, whose label is updated as the
 * window opens/closes.
 *
 */

class WindowHandler extends WindowAdapter implements ActionListener, ComponentListener
{
	Window window;
	Button button;

	//
	// Constructor. Simply call with a window and an associated button.
	//
	public WindowHandler(Window w, Button b)
	{
		window=w;
		button=b;
		w.addWindowListener(this);
		w.addComponentListener(this);
		b.addActionListener(this);
	}

	//
	// Get methods.
	//
	public Window getWindow()	{ return window; }
	public Button getButton()	{ return button; }

	//
	// Action performed == button pressed.
	//
	public void actionPerformed(ActionEvent e)
	{
		if (window.isVisible())
			window.setVisible(false);
		else
			window.show();
	}

	//
	// Window closing event means we want to hide the window.
	//
	public void windowClosing(WindowEvent e)
	{
		window.setVisible(false);
	}

	//
	// When the window is hidden/shown, we want to change the button label accordingly.
	//
	public void componentHidden(ComponentEvent e)
	{
		button.setLabel("Show" + button.getLabel().substring(4));
	}
	public void componentShown(ComponentEvent e)
	{
		button.setLabel("Hide" + button.getLabel().substring(4));
	}

	//
	// Extra methods required by ComponentListener.
	//
	public void componentMoved(ComponentEvent e) {}
	public void componentResized(ComponentEvent e) {}

}

class ConnectionFrame extends Frame implements EventListener, ActionListener
{
	Panel panel;
	GridLayout panelLayout;

	TextField hostnameField;
	TextField portField;

	Button connectButton;
	Button disconnectButton;

	TextArea statusLine;

	RemoteProfiler profiler;

	ProfileUpdater profUpdater;
	Thread profUpdaterThread;

	public void processEvent(EventObject eo)
	{
		ProfileEvent e = (ProfileEvent)eo;
		switch (e.getType()) {
		case ProfileEvent.ATTEMPTING_TO_CONNECT:
			message("\nTrying to connect to " + e.getMessage() + "...");
			break;
		case ProfileEvent.CONNECT_SUCCEEDED:
			connectButton.setEnabled(false);
			disconnectButton.setEnabled(true);
			message("successful.");
			break;
		case ProfileEvent.CONNECT_FAILED:
			message("Error! " + e.getMessage());
			break;
		case ProfileEvent.DISCONNECT:
			disconnectButton.setEnabled(false);
			connectButton.setEnabled(true);
			message("\nDisconnected.");
			break;
		case ProfileEvent.IO_ERROR:
			message("\nI/O error occurred!");
			break;
		case ProfileEvent.GET_DATA:
//			message("\nReceiving data: ");
			break;
		case ProfileEvent.GOT_DATA:
//			message("received "+e.getValue()+" points");
			break;
		case ProfileEvent.BEGIN_ANALYSIS:
//			message("\nBeginning analysis...");
			break;
		case ProfileEvent.ANALYSIS_DATA:
			break;
		case ProfileEvent.DONE_ANALYSIS:
//			message("Finished analysis.");
			break;
		default:
			// something else...
		}
	}

	void message(String s)
	{
		statusLine.append(s);
	}

	public void actionPerformed(ActionEvent e)
	{
		if (e.getSource() == connectButton) {
			try {
				String portText = portField.getText();
				int portnum = Integer.parseInt(portText);
				profiler.setPort(portnum);
				String hostText = hostnameField.getText();
				profiler.setHost(hostText);

				profUpdaterThread = new Thread(profUpdater);
				profUpdaterThread.start();
			}
			catch (NumberFormatException x) { message("Error! Bad port number!\n"); }
		} else if (e.getSource() == disconnectButton) {
			profUpdaterThread.stop();
		} else System.err.println("Error in ButtonListener! Button not recognized.");
	}

	public ConnectionFrame(RemoteProfiler profiler, ProfileUpdater profUpdater)
	{
		super("Remote profile connector");
		this.profiler = profiler;
		this.profUpdater = profUpdater;

		panelLayout = new GridLayout(4, 2, 8, 8);
		panel = new Panel (panelLayout);
		
		Text hostText = new Text(" Machine name:");
		panel.add(hostText);
		
		String hostname = "";
		try { hostname = InetAddress.getLocalHost().getHostName(); }
		catch (UnknownHostException e) { System.err.println("Can't find local host name"); }

		hostnameField = new TextField(hostname);
		panel.add(hostnameField);

		Text portText = new Text(" Port number (TCP):");
		panel.add(portText);

		portField = new TextField("6827", 4);
		panel.add(portField);

		connectButton = new Button("Connect");
		connectButton.setEnabled(true);
		panel.add(connectButton);

		disconnectButton = new Button("Disconnect");
		disconnectButton.setEnabled(false);
		panel.add(disconnectButton);

		connectButton.addActionListener(this);
		disconnectButton.addActionListener(this);

		add("Center",panel);

		statusLine = new TextArea("Ready.", 2, 40, TextArea.SCROLLBARS_VERTICAL_ONLY);
		statusLine.setEditable(false);
		add("South",statusLine);

		pack();
	}

	public void destroy()
	{
		if ((profUpdaterThread != null) && profUpdaterThread.isAlive()) {
			profUpdaterThread.stop();
		}
	}

}

class SliceStatsPanel extends Container
{
	Text funcName;
	Text funcCount;

	//
	// Constructor.
	//
	public SliceStatsPanel()
	{
		// basic layout is 2x1 grid.
		setLayout(new GridLayout(2, 1));

		// top is function name container
		Container funcNameContainer = new Panel();
		funcNameContainer.setLayout(new FlowLayout(FlowLayout.LEFT));
		funcNameContainer.add(new Text(" Name:"));
		funcName = new Text();
		funcNameContainer.add(funcName);

		// bottom is count container
		Container countContainer = new Panel();
		countContainer.setLayout(new FlowLayout(FlowLayout.LEFT));
		countContainer.add(new Text(" Count:"));
		funcCount = new Text();
		countContainer.add(funcCount);

		// add them.
		add(funcNameContainer);
		add(countContainer);
	}

	//
	// Paint method.
	//
	public void paint(Graphics g)
	{
		Dimension size = getSize();
		g.draw3DRect(1, 1, size.width-2, size.height-2, true);
		g.translate(6, 6);
		super.paint(g);
	}

	//
	// Set/clear methods.
	//
	public void set(Slice s)
	{
		funcName.set(s.getLabel());
		funcCount.set(Integer.toString(s.getValue()));
	}
	public void clear()
	{
		funcName.set("");
		funcCount.set("");
	}

	//
	// Size method.
	//
	public Dimension getPreferredSize()
	{
		return new Dimension(250, (getFontMetrics(funcName.getFont()).getHeight()<<1) + 16 + 6);
	}

}

/*
 * ProfileApplet
 *
 * Applet that dynamically displays profile data.
 *
 */
public class ProfileApplet extends Applet
{
	ConnectionFrame cFrame;
	Frame			javaPieFrame;
	Frame			vmPieFrame;
	Frame			optionsFrame;

	FunctionButtonPanel	buttonPanel;

	RemoteProfiler profiler;
	ProfileUpdater profUpdater;

	public void init()
	{
		// set the layout, make a double-buffer panel in the middle.
		setLayout(new BorderLayout());
		DoubleBufferPanel doubleBufferPanel = new DoubleBufferPanel();
		doubleBufferPanel.setLayout(new BorderLayout());
		add("Center", doubleBufferPanel);

		// put a pretty picture panel in the double-buffer panel.
		Image image = getImage(getCodeBase(), "images/orb.gif");
		PrettyPanel pp = new PrettyPanel(image);
		pp.setLayout(new FlowLayout());
		doubleBufferPanel.add("Center", pp);

		// put the overall piechart in there.
		OverallPieChart overallPie = new OverallPieChart();
		pp.add(overallPie);

		// create profiler object and profileupdater object
		profiler = new RemoteProfiler();
		profUpdater = new ProfileUpdater(profiler);

		// create connection window
		cFrame = new ConnectionFrame(profiler, profUpdater);
		cFrame.setLocation(258, 0);
		cFrame.show();

		// create java piechart
		PieChart javaPieChart = new PieChart("Java Piechart", false);

		// create the updater for the java piechart
		JavaUpdater javaUpdater = new JavaUpdater();
		javaUpdater.setPieChart(javaPieChart);

		// create the frame for the java piechart
		javaPieFrame = new PieChartFrame(javaPieChart, javaUpdater);
		javaPieFrame.setLocation(260, 190);

		// create VM piechart.
		PieChart vmPieChart = new PieChart("VM", false);
		
		// create the updater for the VM piechart
		VMUpdater vmUpdater = new VMUpdater();
		vmUpdater.setPieChart(vmPieChart);

		// create the frame for the VM piechart.
		vmPieFrame = new PieChartFrame(vmPieChart, vmUpdater);
		vmPieFrame.setLocation(516, 190);

		// create options window.
		optionsFrame = new Frame("Options");

		// put button panel in the main frame to the south.
		buttonPanel = new FunctionButtonPanel(cFrame, profiler, javaPieFrame, vmPieFrame, optionsFrame, this);
		add("South", buttonPanel);

		// add all the listeners to the profiler updater.
		profUpdater.addListener(cFrame);
		profUpdater.addListener(overallPie);
		profUpdater.addJavaListener(javaUpdater);
		profUpdater.addVMListener(vmUpdater);

		setSize(250, 400);
	}

	public void destroy()
	{
		if (cFrame != null) cFrame.destroy();
	}

}

/*
 * FunctionButtonPanel
 *
 * Creates and handles buttons in the function button panel.
 *
 */
class FunctionButtonPanel extends Container implements ActionListener
{

	// window handling buttons
	WindowHandler connect;
	WindowHandler javaPie;
	WindowHandler vmPie;
	WindowHandler options;

	// function buttons
	Button clear_b;
	Button quit_b;

	// profiler object.
	Profiler prof;

	// applet thread
	Applet applet;

	FunctionButtonPanel(Window cf, Profiler p, Window jp, Window vp, Window o, Applet a)
	{
		prof = p;
		applet = a;
		setLayout(new GridLayout(3, 2, 5, 5));

		// create the buttons
		Button	connect_b	=	new Button("Hide connect window");
				clear_b		=	new Button("Clear data");
		Button	javaPie_b	=	new Button("Show Java piechart");
		Button	vmPie_b		=	new Button("Show VM piechart");
		Button	options_b	=	new Button("Show options window");
				quit_b		=	new Button("Stop");

		// add them
		add(connect_b);
		add(clear_b);
		add(javaPie_b);
		add(vmPie_b);
		add(options_b);
		add(quit_b);
	
		// create the WindowHandler objects
		connect		=	new WindowHandler(cf, connect_b);
		javaPie		=	new WindowHandler(jp, javaPie_b);
		vmPie		=	new WindowHandler(vp, vmPie_b);
		options		=	new WindowHandler(o, options_b);

		// listen to the other buttons
		clear_b.addActionListener(this);
		quit_b.addActionListener(this);
	}

	public void actionPerformed(ActionEvent e)
	{
		Button b = (Button) e.getSource();
		if (b == clear_b) {
			// clear the data
			try { prof.reset(); } catch (IOException x) {}
			return;
		}
		if (b == quit_b) {
			// stop.
			return;
		}
	}
}

/*
 * PieChartFrame
 *
 * Frame to hold a piechart.
 *
 */
class PieChartFrame extends Frame
{

	PieChart pc;				// piechart in this frame
	SliceList list;				// list of slices
	SliceStatsPanel ssPanel;	// slice stats panel.
	JavaUpdater jupdater;		// optional java updater.
	VMUpdater vupdater;			// optional VM updater.
	PopupMenu pp;				// optional popup menu.
	MouseHandler sh;			// handler for mouse events.
	MenuHandler mh;				// handler for menu events.

	Button clearData;
	Button hidePie;
	Button hideList;
	Button refresh;

	//
	// Constructors.
	//
	// They create the frame, but don't show it.
	//
	public PieChartFrame(PieChart pc, VMUpdater vup)
	{
		this(pc); setVMUpdater(vup);
	}
	public PieChartFrame(PieChart pc, JavaUpdater jup)
	{
		this(pc); setJavaUpdater(jup);
	}
	public PieChartFrame(PieChart pc)
	{
		super(pc.getName());
		this.pc = pc;

		setLayout(new BorderLayout());

		// create double-buffered panel and add it to the frame.
		DoubleBufferPanel db = new DoubleBufferPanel();
		add("Center", db);

		// add the piechart to the double-buffered panel.
		db.add(pc);

		// create slice stats panel, and add it to the frame.
		ssPanel = new SliceStatsPanel();
		add("South", ssPanel);

		// create a button panel and add it to the frame.
		Panel buttonPanel = new Panel();
		buttonPanel.setLayout(new GridLayout(1, 5));
		add("North", buttonPanel);

		// create "hide window" button and add it to the button panel.
		Button closeb = new Button("Hide window");
		buttonPanel.add(closeb);

		// create "clear data" button and add it to the button panel.
		clearData = new Button("Clear data");
		buttonPanel.add(clearData);

		// create "hide piechart" button and add it to the button panel.
		hidePie = new Button("Hide piechart");
		buttonPanel.add(hidePie);

		// create "hide list" button and add it to the button panel.
		hideList = new Button("Hide list");
		buttonPanel.add(hideList);

		// create "refresh" button and add it to the button panel.
		refresh = new Button("Refresh");
		buttonPanel.add(refresh);

		// create window handler.
		WindowHandler wh = new WindowHandler(this, closeb);

		//methodFrame.setLocation(520, 200);

		// create slice list and register it with the piechart.
		list = new SliceList(pc);
		pc.setList(list);
		add("East", list);

		// create mouse handler and register it with the piechart.
		sh = new MouseHandler();
		pc.addMouseListener(sh);

		// create button handler and register it with the buttons.
		ButtonHandler bh = new ButtonHandler();
		clearData.addActionListener(bh);
		hidePie.addActionListener(bh);
		hideList.addActionListener(bh);
		refresh.addActionListener(bh);

		// pack frame.
		pack();
	}

	public PieChart getPieChart()	{ return pc; }

	public void setJavaUpdater(JavaUpdater jup)
	{
		jupdater = jup;

		// if we have a java updater, we also should have a popup menu..
		if (pp == null) {
			pp = new PopupMenu();
			
			MenuItem winItem = new MenuItem("Display window");
			winItem.setActionCommand("w");
			pp.add(winItem);

			MenuItem sItem = new MenuItem("Display source");
			sItem.setActionCommand("s");
			pp.add(sItem);

			MenuItem bItem = new MenuItem("Display bytecodes");
			bItem.setActionCommand("b");
			pp.add(bItem);

			MenuItem mItem = new MenuItem("Display machine code");
			mItem.setActionCommand("m");
			pp.add(mItem);

			mh = new MenuHandler();
			pp.addActionListener(mh);
		}
	}
	
	public void setVMUpdater(VMUpdater vup)
	{
		vupdater = vup;
	}

	//
	// Required for MouseListener.
	//
	public void mousePressed(MouseEvent e) {}
	public void mouseReleased(MouseEvent e)	{}

	//
	// Internal method to show the window of a given slice.
	//
	private void showPieChart(Slice s)
	{
		// show piechart.
		if (jupdater == null) return;
		JavaMethod jm = jupdater.getJavaMethodFromSlice(s);
		if (jm != null) {
			Window w = jm.getWindow();
			if (w == null) {
				// window not yet created ... let's create it.
				w = new PieChartFrame(jm.getPieChart());
				jm.setWindow(w);
			}
			// show the window.
			w.show();
		}
	}

	//
	// Internal class to handle mouse events.
	//
	class MouseHandler extends MouseAdapter
	{
		Slice showing;

		//
		// When the mouse is clicked, open a menu or something ...
		//
		public void mouseClicked(MouseEvent e)
		{
			if (e.isShiftDown()) {
				// show piechart.
				showPieChart((Slice) e.getComponent());
			} else 
			if (e.isControlDown()) {

			} else 
			if (e.isAltDown()) {

			} else {
				// show menu.
				if (showing == null) return;	// we shouldn't be getting click events before
												// enter events ...
				if (pp != null) {
					showing.add(pp);	// add(popup) also removes it from its old parent.
					pp.show(e.getComponent(), e.getX(), e.getY());
				}
			}
		}
		//
		// When the mouse enters a slice, highlight it and display the info.
		//
		public void mouseEntered(MouseEvent e)
		{
			showing = (Slice)e.getComponent();
			showing.setSelected(true);
			if (ssPanel != null) ssPanel.set(showing);
			list.drawSlice(showing);
		}
		//
		// When the mouse exits a slice, de-highlight it and clear the info.
		//
		public void mouseExited(MouseEvent e)
		{
			showing = null;
			Slice s = (Slice)e.getComponent();
			s.setSelected(false);
			if (ssPanel != null) ssPanel.clear();
			list.drawSlice(s);
		}
	}

	//
	// Internal class to handle menu events.
	//
	class MenuHandler implements ActionListener
	{
		public void actionPerformed(ActionEvent e)
		{
			if (e.getActionCommand().equals("w")) {
				// open window.
				Menu m = (Menu) e.getSource();
				Slice s = (Slice) m.getParent();
				showPieChart(s);

			} else if (e.getActionCommand().equals("s")) {
				// display source code listing

			} else if (e.getActionCommand().equals("b")) {
				// display bytecode listing

			} else if (e.getActionCommand().equals("m")) {
				// display generated machine code listing

			} else {}
		}
	}

	//
	// Internal class to handle button events.
	//
	class ButtonHandler implements ActionListener
	{
		public void actionPerformed(ActionEvent e)
		{
			if (e.getSource()==clearData) {

				if (vupdater != null) {
					// VM
					pc.removeAllSlices();
					vupdater.clearAllData();
				} else if (jupdater != null) {
					// java
					pc.removeAllSlices();
					jupdater.clearAllData();
				} else {
					// line number
					pc.removeAllSlices();
				}

				// now repaint the piechart.
				pc.repaint();

			} else if (e.getSource()==hidePie) {
				// hide/show the piechart and change the button label.
				if (pc.isVisible()) {
					pc.setVisible(false);
					hidePie.setLabel("Show piechart");
				} else {
					pc.setVisible(true);
					hidePie.setLabel("Hide piechart");
				}

			} else if (e.getSource()==hideList) {
				// hide/show the list and change the button label.
				if (list.isVisible()) {
					list.setVisible(false);
					hideList.setLabel("Show list");
				} else {
					list.setVisible(true);
					hideList.setLabel("Hide list");
				}

			} else if (e.getSource()==refresh) {
				// refresh the piechart
				pc.repaint();

			}
		}
	}

}

/*
 * OverallPieChart
 *
 * Creates, displays, and updates the overall piechart.
 *
 */
class OverallPieChart extends PieChart implements EventListener, ActionListener
{
	Slice nativeSlice;
	Slice VMSlice;
	Slice javaSlice;

	WindowHandler javaHandler;
	WindowHandler VMHandler;

	OverallPieChart()
	{
		super("Overall");
		nativeSlice = super.addSlice("Native", 0, Color.red);
		VMSlice = super.addSlice("VM internal", 0, Color.blue);
		javaSlice = super.addSlice("Java", 0, Color.green);
		
		minAngle = 0;

		addActionListener(this);
	}

	public void actionPerformed(ActionEvent e)
	{
		Slice s = (Slice) e.getSource();
		if (s == javaSlice) {
			System.out.println("Java slice clicked");

		} else
		if (s == VMSlice) {
			System.out.println("VM slice clicked");

		} else
		if (s == nativeSlice) {
			System.out.println("Native slice clicked");
			// nothing (yet...)

		} else {}
	}

	public void processEvent(EventObject eo)
	{
		ProfileEvent e = (ProfileEvent)eo;
		// wait for the done-analysis event.
		if (e.getType() != ProfileEvent.DONE_ANALYSIS) return;

		//System.out.println("Updating overall pie...");

		ProfileUpdater pu = (ProfileUpdater)e.getSource();

		try {
		super.disablePaint();
		super.changeSlice(nativeSlice, pu.profiler.getNative());
		super.changeSlice(VMSlice, pu.profiler.getVM());
		super.changeSlice(javaSlice, pu.profiler.getJava());
		super.enablePaint();
		repaint();
		} catch (IOException x) {}
	}


}

/*
 * PrettyPanel
 *
 * Gives us a static background picture.
 *
 */
class PrettyPanel extends Container implements ImageObserver
{
	Image background;
	
	PrettyPanel(Image background)
	{
		this.background = background;
	}

	/*
	 * override update to *not* erase the background before painting
	 */
	public void update(Graphics g) { paint(g); }
	
	/*
	 * paint the background picture, then call super.paint which
	 * will paint all contained components 
	 */
	public void paint(Graphics g)
	{
		// paint the background image
		g.drawImage(background, 0, 0, getSize().width, getSize().height,
			this);
		super.paint(g);
	}
}
