import java.awt.AWTEvent;				// for constructor
import java.awt.AWTEventMulticaster;	// for addActionListener, removeActionListener
import java.awt.Color;					// for color, constructor, getColor, setColor, paint
import java.awt.Component;				// for superclass
import java.awt.Dimension;				// for paint, contains
import java.awt.event.ActionEvent;		// for processMouseEvent
import java.awt.event.ActionListener;	// for actionListener, addActionListener, removeActionListener, processMouseEvent
import java.awt.event.MouseEvent;		// for processMouseEvent
import java.awt.Graphics;				// for update, paint
import java.awt.Point;					// for contains
import java.awt.FontMetrics;			// for drawLabel

	/*
	 *
	 * Slice
	 * Implements a slice of a PieChart.
	 * 360 degree accuracy. Uses Graphics.fillArc for drawing.
	 *
	 */
	class Slice extends Component
	{

		ActionListener actionListener;	// Post action events (clicks) to listeners.

		int value;			// value
		float startAngle;	// offset in degrees
		float arcAngle;		// value of this slice
		Color color;		// color to draw this slice
		int index;			// which slice number in the piechart
		float minAngle =15f;// minimum angle for which we still want to put a label.
		boolean drawLabel;	// whether we want to display label or not.
		Color labelColor;	// color for the (optional) label.
		boolean selected;	// whether or not this slice is selected.

		//
		// Constructors
		// 
		// (don't you wish Java had default values for arguments?)
		public Slice()						{ this("", 0, Color.white);		}
		public Slice(String label)			{ this(label, 0, Color.white);	}
		public Slice(int value)				{ this("", value, Color.white);	}
		public Slice(Color c)				{ this("", 0, c);				}
		public Slice(String label, Color c)	{ this(label, 0, c);			}
		public Slice(int value, Color c)	{ this("", value, c);			}
		public Slice(String label, int value)	{ this(label, value, Color.white); }
		public Slice(String label, int value, Color color)
		{
			setName(label);
			this.value = value;
			this.color = color;
			enableEvents(AWTEvent.MOUSE_EVENT_MASK);	// mouse events only
		}

		//
		// Methods to get/set instance variables
		//
		public String getLabel()				{ return super.getName();		}
		public void setLabel(String label)		{ super.setName(label);			}
		public String getName()					{ return super.getName()+":"+value;	}
		public Color getLabelColor()			{ return labelColor;			}
		public void setLabelColor(Color color)	{ labelColor = color;			}
		public int getValue()					{ return value;					}
		public void setValue(int value)			{ this.value = value;			}
		public void incValue()					{ ++this.value;					}
		public int getIndex()					{ return index;					}
		public void setIndex(int index)			{ this.index = index;			}
		public void setDrawLabel(boolean b)		{ drawLabel = b;				}
		public float getMinAngle()				{ return minAngle;				}
		public void setMinAngle(float min)		{ minAngle = min;				}
		public Color getColor()					{ return color;					}
		public void setColor(Color color)
		{
			this.color = color;
			//repaint(); // no need to invalidate region, too inefficent!
			// todo: find out why this paint call causes the entire
			// piechart to be repainted ... (and its not super.paint())
			paint(getGraphics());
		}
		public boolean getSelected()			{ return selected;				}
		public void setSelected(boolean selected)
		{
			this.selected = selected;
			//repaint(); // no need to invalidate region, too inefficent!
			// todo: find out why this paint call causes the entire
			// piechart to be repainted ... (and its not super.paint())
			paint(getGraphics());
		}
		public void setAngles(float startAngle, float arcAngle)
		{
			this.startAngle = startAngle;
			this.arcAngle = arcAngle;
		}

		//
		// Paint method.
		//
		public void paint(Graphics g) 
		{
			if (g == null) {
				System.err.println("Bug in Java! Bad graphics context.");
				return;
			}
			Dimension size = getSize();
			if (selected) g.setColor(color.brighter());
			else g.setColor(color);
			g.fillArc(0, 0, size.width, size.height, (int)startAngle, (int)arcAngle);
			if (drawLabel && (arcAngle > minAngle)) {
				g.setColor(color.darker());
				drawLabel(g);
			}
			super.paint(g);
		}

		//
		// Method to draw the slice label.
		//
		public void drawLabel(Graphics g)
		{
			// figure out where the label should go.
			double angle = startAngle / 180. * Math.PI + arcAngle / 360. * Math.PI;
			Dimension size = getSize();
			int cx = size.width >> 1;
			int cy = size.height >> 1;
			int xpos = cx + (int)((Math.cos(angle) * cx) * .6);
			int ypos = cy - (int)((Math.sin(angle) * cy) * .6);

			// figure out where to put the text such that its center is at (xpos, ypos)
			String label = super.getName();
			FontMetrics fm = g.getFontMetrics();
			ypos += fm.getHeight() >> 1;
			xpos -= fm.stringWidth(label) >> 1;

			// finally, paint the text.
			g.drawString(label, xpos, ypos);
		}

		//
		// Method which returns whether the point (x,y) is in this slice.
		//
		public boolean contains(Point p)	{ return contains(p.x, p.y);	}
		public boolean contains(int x, int y)
		{
			// find what angle the point is at.
			Dimension size = getSize();
			int cx = size.width >> 1;
			int cy = size.height >> 1;
			float angle = 180f+(float)(180 / Math.PI *
				Math.atan2((double)(y - cy), (double)(cx - x)));
			
			if ((angle<startAngle) || (angle>startAngle+arcAngle)) return false;
			
			// don't have to check if it is in radius because the PieChart handles that.
			return true;
			// return (((x-cx)*(x-cx) + (y-cy)*(y-cy)) <= cx*cx);
		}
		
		//
		// Methods to add/remove listeners.
		//
		public void addActionListener(ActionListener listener)
		{
			actionListener = AWTEventMulticaster.add(actionListener, listener);
		}
		public void removeActionListener(ActionListener listener)
		{
			actionListener = AWTEventMulticaster.remove(actionListener, listener);
		}

		//
		// Methods to handle mouse events.
		//
		public void processMouseEvent(MouseEvent e)
		{
			if (e.getID() == MouseEvent.MOUSE_CLICKED) {
				// dispatch this to all listeners
				if(actionListener != null)
					actionListener.actionPerformed(new ActionEvent(
						this, ActionEvent.ACTION_PERFORMED, super.getName(), e.getModifiers()));
			}
			super.processMouseEvent(e);
		}
	}
