import java.awt.*;
import java.awt.event.*;

/*
 *
 * SliceList
 *
 * This lightweight component displays the slices of a piechart in a list.
 * Calling its repaint() method will refresh it. It also forwards mouse events.
 *
 *
 */
public class SliceList extends Component
{
	PieChart watch;					// piechart to watch.
	private int currentLocation;	// current location (y coordinate)
	private int fontHeight;			// cached font height (updated on painting, size calls)

	private static final float RATIO = 0.8f;	// ratio of space for name to space for size.
	private static final Color TITLE_COLOR = Color.black;		// color of title.
	private static final Color SEL_BACK_COLOR = Color.black;	// background for selection.
	private static final Color SEL_FORE_COLOR = Color.white;	// foreground for selection.

	//
	// Constructors.
	//
	public SliceList(PieChart c)
	{
		watch = c; fontHeight = 15;
		MouseHelper mh = new MouseHelper();
		addMouseListener(mh);
		addMouseMotionListener(mh);
	}

	//
	// Paint method.
	//
	public synchronized void paint(Graphics g)
	{
		Dimension size = getSize();

		// Create two smaller graphics contexts, so we can draw on each side seperately.
		int nwidth = (int)(size.width*RATIO);
		Graphics g1 = g.create(0, 0, nwidth, size.height);
		Graphics g2 = g.create(nwidth, 0, size.width - nwidth, size.height);

		// update font height
		fontHeight = g.getFontMetrics().getHeight();

		Component[] c = watch.getComponents();	// this makes a copy and is synch, so is safe.
		currentLocation = 0;
		Rectangle clipRect = g.getClipBounds();
		int max = clipRect.height + clipRect.y;

		// draw title
		g1.setColor(TITLE_COLOR);
		g2.setColor(TITLE_COLOR);
		currentLocation += g.getFontMetrics().getHeight();
		g1.drawString("Name", 0, currentLocation);
		g2.drawString("Value", 0, currentLocation);

		// draw list
		for (int i=1; i<c.length; ++i) {	// start at 1 to skip "other" slice.
			Slice s = (Slice) c[i];
			if (write(g1, g2, s, max, false)) break;
		}
		//write(g1, g2, (Slice)c[0], max); // "other" slice goes at the end.

		super.paint(g);
	}

	//
	// Draws only the specified slice.
	//
	public synchronized void drawSlice(Slice s)
	{
		if (s.getIndex() == 0) return; // don't draw other slice over title!
		Dimension size = getSize();
		int nwidth = (int)(size.width*RATIO);
		Graphics g = getGraphics();
		Graphics g1 = g.create(0, 0, nwidth, size.height);
		Graphics g2 = g.create(nwidth, 0, size.width - nwidth, size.height);
		currentLocation = s.getIndex() * fontHeight;
		write(g1, g2, s, 0, true);
	}

	//
	// Writes the slice s at the current location on graphics context g1/g2.
	// Returns whether the new current location is greater than max_y.
	//
	private boolean write(Graphics g1, Graphics g2, Slice s, int max_y, boolean clear)
	{
		if (s.getSelected()) {
			int descent = g1.getFontMetrics().getDescent();
			g1.setColor(SEL_BACK_COLOR);
			g2.setColor(SEL_BACK_COLOR);
			g1.fillRect(0, currentLocation + descent, g1.getClipBounds().width, fontHeight);
			g2.fillRect(0, currentLocation + descent, g2.getClipBounds().width, fontHeight);
			g1.setColor(SEL_FORE_COLOR);
			g2.setColor(SEL_FORE_COLOR);
		} else {
			if (clear) {
				int descent = g1.getFontMetrics().getDescent();
				g1.clearRect(0, currentLocation + descent, g1.getClipBounds().width, fontHeight);
				g2.clearRect(0, currentLocation + descent, g2.getClipBounds().width, fontHeight);
			}
			g1.setColor(s.getColor());
			g2.setColor(s.getColor());
		}
		currentLocation += fontHeight;
		//System.out.println(currentLocation+": "+s);
		g1.drawString(s.getLabel(), 0, currentLocation);
		g2.drawString(Integer.toString(s.getValue()), 0, currentLocation);
		return (currentLocation >= max_y);
	}

	//
	// Size methods.
	//
	public Dimension getPreferredSize()
	{
		Graphics g = getGraphics();
		if (g != null) fontHeight = g.getFontMetrics().getHeight();
		return new Dimension(250, 6 * fontHeight);
	}
	public Dimension getMaximumSize()
	{
		Graphics g = getGraphics();
		if (g != null) fontHeight = g.getFontMetrics().getHeight();
		return new Dimension(Integer.MAX_VALUE, watch.getComponentCount() * fontHeight);
	}

	class MouseHelper extends MouseAdapter implements MouseMotionListener
	{
		Slice showing = null;

		public void mouseExited(MouseEvent e)
		{
			//System.out.println("Mouse exited: "+e.getX()+","+e.getY());
			if (showing != null) {
				showing.dispatchEvent(new MouseEvent(showing, MouseEvent.MOUSE_EXITED,
					e.getWhen(), e.getModifiers(), e.getX(), e.getY(), e.getClickCount(),
					e.isPopupTrigger()));
				showing = null;
			}
		}
		public void mouseClicked(MouseEvent e)
		{
			//System.out.println("Mouse clicked: "+e.getX()+","+e.getY());
			if (showing != null) {
				showing.dispatchEvent(new MouseEvent(SliceList.this, MouseEvent.MOUSE_CLICKED,
					e.getWhen(), e.getModifiers(), e.getX(), e.getY(), e.getClickCount(),
					e.isPopupTrigger()));
			}
		}
		public void mouseDragged(MouseEvent e)
		{
			mouseMoved(e);
		}
		public void mouseMoved(MouseEvent e)
		{
			//System.out.println("Mouse moved: "+e.getX()+","+e.getY());
			int element_num = e.getY() / fontHeight;
			if (element_num < 1) return; // ignore title.
			Slice now_showing = (Slice) watch.getSlice(element_num);
			if (now_showing == showing) return;
			if (showing != null)
				showing.dispatchEvent(new MouseEvent(showing, MouseEvent.MOUSE_EXITED,
					e.getWhen(), e.getModifiers(), e.getX(), e.getY(), e.getClickCount(),
					e.isPopupTrigger()));
			showing = now_showing;
			if (showing != null)
				showing.dispatchEvent(new MouseEvent(showing, MouseEvent.MOUSE_ENTERED,
					e.getWhen(), e.getModifiers(), e.getX(), e.getY(), e.getClickCount(),
					e.isPopupTrigger()));

		}
	}

}
