//
// TransCtl.cpp : Implementation of CTransCtl
//
// Transparent Control OCX
// Tom Armstrong & Mark Nelson
// September 1997
//
// This file contains much of the implementation
// for the Transparent control.  The rest of the
// implementation is in the companion header file.
//

#include "stdafx.h"
#include "TransparentControl.h"

#include "dib.h"
#include "BindStatusCallback.h"
#include "TransCtl.h"

//////////////////////
// CTransCtl
//////////////////////
//
// The ImageFile property should be the complete
// path of a BMP file.  At this time we're not
// implementing any palette manipulation, so
// systems with 256 color video displays should
// stick with the system colors when using this
// control.
//
STDMETHODIMP CTransCtl::get_ImageFile(BSTR * pVal)
{
  *pVal = m_bstrImageFile.Copy();
  return S_OK;
}

STDMETHODIMP CTransCtl::put_ImageFile(BSTR newVal)
{
  m_bstrImageFile = newVal;
  //
  // Only perform the download at runtime
  //
  BOOL bUserMode;
  GetAmbientUserMode( bUserMode );
  if ( bUserMode ) {
    //
    // Inform the control that it can no longer
    // draw the old image
    //
    put_ReadyState( READYSTATE_LOADED );
    //
    // Reinitialize the DIB
    //
    m_dib.Reset();
    //
    // Start the asynchronous download
    //
    COurBindStatusCallback<CTransCtl>::Download( this,
                                                 OnData,
                                                 m_bstrImageFile,
                                                 m_spClientSite,
                                                 FALSE );
  }
  return S_OK;
}

//
// This routine is called to draw a text string in the control.
// When there isn't any picture ready to draw we use this
// routine to try to breath some life into the control.
//
void CTransCtl::DrawString( HDC hdc, TCHAR* szMsg, RECT* rc )
{
  DrawText( hdc,
            szMsg,
            lstrlen( szMsg ),
            rc,
            DT_WORDBREAK );
}

//
// This is where all of the work needed to draw a control
// with random transparent areas takes place.
//
HRESULT CTransCtl::OnDraw( ATL_DRAWINFO& di )
{
  USES_CONVERSION;
  BOOL bRet;

  //
  // Creating this clipping region prevents me from coloring
  // outside the lines if my bitmap size doesn't match up properly
  // with what I'm being asked to draw.
  //
  RECT& rc = *(RECT*)di.prcBounds;
  HRGN region = ::CreateRectRgnIndirect( &rc );
  ::SelectClipRgn( di.hdcDraw , region );
  //
  // If we're drawing in design mode, just
  // fill the rectangle with a white background
  //
  BOOL bUserMode;
  GetAmbientUserMode( bUserMode );
  if (! bUserMode ) {
    USES_CONVERSION;
    HDC hdc = di.hdcDraw;
    RECT &rc = * (RECT*) di.prcBounds;

    FillRect( hdc,
              &rc,
              (HBRUSH) GetStockObject( WHITE_BRUSH ));
    //
    // Now draw the ambient display name
    //
    BSTR bstr;
    if ( SUCCEEDED( GetAmbientDisplayName( bstr )) ) {
      USES_CONVERSION;
      DrawText( hdc,
                OLE2A( bstr ),
                -1,
                &rc,
                DT_TOP | DT_SINGLELINE );
    }
    //
    // At design time, we don't try to download
    // a remote image as it will not work in
    // most containers.
    //
    return S_OK;
  }
  //
  // If we're still downloading the image,
  // draw some text and return
  //
  if ( m_nReadyState != READYSTATE_COMPLETE ) {
    DrawString( di.hdcDraw,
                "Downloading bitmap...",
                &rc );
    return S_OK;
  }
  //
  // If this is the first time through OnDraw,
  // we need to setup the DIB
  //
  if ( !m_dib.IsInitialized() ) {
    m_dib.SetBitmapInfoHeader( m_pbBuffer );
    m_dib.Create( di.hdcDraw );
    m_dib.SetBits( m_pbBuffer );
  }

  int rows = m_dib.GetHeight();
  int columns = m_dib.GetWidth();
  //
  // This code gets the bitmap object, creates a compatible DC for it,
  // and selects it into that DC.
  //
  // I'm also going to create a mask bitmap that will be used to define
  // the transparent areas.  This is best done by creating a monochrome
  // bitmap.  If I set its background color to be the same color as the
  // transparent color in the bitmap, a copy into it will set all transparent
  // bits to 1, and all others to 0.
  //
  HDC hdcMask = ::CreateCompatibleDC( di.hdcDraw );
  HBITMAP bmMask = ::CreateBitmap( columns, rows, 1, 1, NULL );
  HBITMAP hOldMaskBitmap = (HBITMAP) SelectObject( hdcMask, bmMask );

  SetBkColor( HDC( m_dib ), RGB( 255, 255, 255 ) );
  SetTextColor( HDC( m_dib ), RGB( 0, 0, 0 ) );
  bRet = BitBlt( hdcMask,
                 0,
                 0,
                 columns,
                 rows,
                 HDC( m_dib ),
                 0,
                 0,
                 SRCCOPY );
  if ( 0 == bRet ) {
    DWORD error = GetLastError();
    FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM,
                   0,
                   error,
                   0,
                   m_strError,
                   256,
                   0 );
    DrawString( di.hdcDraw,
                m_strError,
                &rc );
    return S_OK;
  }

  //
  // In order for the masking to work properly, the destination dc has
  // to have a background color of white and a text color of black.
  // I change these back to their original values at the end of the
  // routine
  //
  COLORREF crOldBackColor = ::GetBkColor( di.hdcDraw );
  COLORREF crOldTextColor = ::GetTextColor( di.hdcDraw );
  ::SetBkColor( di.hdcDraw, RGB( 0xff, 0xff, 0xff ) );
  ::SetTextColor( di.hdcDraw, RGB( 0, 0, 0 ) );
  //
  // This step XORs the current screen image with our bitmap.  This might
  // seem counterproductive, since it leaves *none* of the pixels in the
  // state we want them.  But... two more steps and everything will be
  // fine.  The next step will separate the opaque and transparent bits by
  // clearing all the opaque bits to solid black, but leaving the transparent
  // bits alone.  The final step will be to apply another XOR.  The transparent
  // bits are then going to be right back where they were when this process
  // started, since XOR( XOR( x ) ) == x.  The opaque bits will be set to
  // the desired color in the bitmap.
  //
  bRet = BitBlt( di.hdcDraw,
                 rc.left,
                 rc.top,
                 columns,
                 rows,
                 HDC( m_dib ),
                 0,
                 0,
                 SRCINVERT );
  if ( 0 == bRet ) {
    DWORD error = GetLastError();
    FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM,
                   0,
                   error,
                   0,
                   m_strError,
                   256,
                   0 );
    DrawString( di.hdcDraw,
                m_strError,
                &rc );
    return S_OK;
  }
  //
  // Now I clear all of the opaque bits by applying a mask.  In the final
  // step, my XOR of the screen with the image takes all of the opaque
  // bits and sets them to their desired state, since (a XOR 0) == a.
  //
  bRet = BitBlt( di.hdcDraw,
                 rc.left,
                 rc.top,
                 columns,
                 rows,
                 hdcMask,
                 0,
                 0,
                 SRCAND );

  if ( 0 == bRet ) {
    DWORD error = GetLastError();
    FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM,
                   0,
                   error,
                   0,
                   m_strError,
                   256,
                   0 );
    DrawString( di.hdcDraw,
                m_strError,
                &rc );
    return S_OK;
  }
  //
  // Finally, we draw the actual image.  All the opaque bits get
  // inverted to where they are supposed to be.  All of the transparent
  // bits are toggled back to the state they were in when we startedd
  // this routine.
  bRet = BitBlt( di.hdcDraw,
                 rc.left,
                 rc.top,
                 columns,
                 rows,
                 HDC( m_dib ),
                 0,
                 0,
                 SRCINVERT );
  if ( 0 == bRet ) {
    DWORD error = GetLastError();
    FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM,
                   0,
                   error,
                   0,
                   m_strError,
                   256,
                   0 );
    DrawString( di.hdcDraw,
                m_strError,
                &rc );
    return S_OK;
  }
  if ( hOldMaskBitmap )
    ::SelectObject( hdcMask, hOldMaskBitmap );
  DeleteDC( hdcMask );
  //
  // restore the DC colors
  //
  ::SetBkColor( di.hdcDraw, crOldBackColor );
  ::SetTextColor( di.hdcDraw, crOldTextColor );
  return S_OK;
}

