//
// Dib.cpp
//
// Transparent Control OCX
// Tom Armstrong & Mark Nelson
// September 1997
//
// This is a utility class used to contain a DIB.
// It provides a little encapsulation to make the
// rest of the project easier to follow.
//

#include "stdafx.h"
#include "Dib.h"

#define WIDTHBYTES(bits)    (((bits) + 31) / 32 * 4)

//
// The ctor doesn't try to create a valid
// DIB, it simply makes sure all the members
// are in a known state.  We won't try to
// render the DIB while it's in this state.
//
CDib::CDib()
{
  m_dwHeaderSize = 0;
  m_dwImageSize = 0;
  m_uBMI.p = 0;
  m_pbBits = 0;
  m_memDC = 0;
  m_hSection = 0;
  m_hOldBitmap = 0;
  m_lHeight = 0;
  m_lWidth = 0;
  m_bInitialized = FALSE;
}
//
// The dtor relies on the Reset() member
// to free up all the resources that may
// have been allocated to hold a DIB
//

CDib::~CDib()
{
  Reset();
}

//
// The CDib object potentially is
// holding onto a block of memory that
// holds the image, a DC that it has
// been rendered on, and a DIBSection
// object.  We have to return all these
// things to the system when we're no
// longer using them, and that's what
// this function does.
//
void CDib::Reset()
{
  if( m_uBMI.p ) {
    delete[] m_uBMI.p;
    m_uBMI.p = 0;
  }
  if( m_memDC ) {
    if( m_hOldBitmap )
      ::SelectObject( m_memDC, m_hOldBitmap );
    ::DeleteDC( m_memDC );
    m_hOldBitmap = 0;
    m_memDC = 0;
  }
  if( m_hSection ) {
    ::DeleteObject( m_hSection );
    m_hSection = 0;
  }
  m_bInitialized = FALSE;
}

//
// Calculating the number of bytes needed to hold the
// image is a little tricky, because the internal
// representation of a DIB has some odd requirements
// regarding the number of bytes used to store a row.
//
DWORD CDib::GetImageSize()
{
  DWORD &dw = m_uBMI.p->bmiHeader.biSizeImage;
  if( dw == 0 ) {
    dw = m_uBMI.p->bmiHeader.biHeight *
         WIDTHBYTES( (DWORD) m_uBMI.p->bmiHeader.biWidth *
                             m_uBMI.p->bmiHeader.biBitCount );
  }
  return( dw );
}

BOOL CDib::SetBitmapInfoHeader( BYTE* buffer )
{
  BITMAPFILEHEADER bmfh;
  memcpy( &bmfh, buffer, sizeof( bmfh ));
  if( bmfh.bfType != 0x4d42 )
    return FALSE;
  m_dwHeaderSize = bmfh.bfOffBits - sizeof( bmfh );
  m_uBMI.pBytes = new BYTE[ m_dwHeaderSize ];
  if( !m_uBMI.pBytes )
    return FALSE;
  memcpy( m_uBMI.pBytes,
          buffer + sizeof( BITMAPFILEHEADER ),
          m_dwHeaderSize );
  return TRUE;
}

//
// Once you've read in the bits and header
// information, this function is used to
// select it into a memory DC where it
// can be quickly blitted whenever you wish.
//
DWORD CDib::Create( HDC hdc )
{
  m_memDC = ::CreateCompatibleDC( hdc );
  if (! m_memDC )
    return ::GetLastError();
  m_hSection = ::CreateDIBSection( m_memDC,
                                   m_uBMI.p,
                                   DIB_RGB_COLORS,
                                   (void **) &m_pbBits,
                                   0,
                                   0 );
  if ( !m_hSection )
    return ::GetLastError();
  m_hOldBitmap = (HBITMAP) ::SelectObject( m_memDC, m_hSection );
  if ( !m_hOldBitmap )
    return ::GetLastError();
  m_lWidth = m_uBMI.p->bmiHeader.biWidth;
  m_lHeight = m_uBMI.p->bmiHeader.biHeight;
  return 0;
}

//
// Copy a block of bits into the DIB
//
void CDib::SetBits( BYTE* buffer )
{
  memcpy( m_pbBits,
          buffer + sizeof( BITMAPFILEHEADER ) + m_dwHeaderSize,
          GetImageSize() );
  m_bInitialized = TRUE;
}

//
// Since we already have the DIB selected
// into a memory context, drawing it is
// simply a matter of doing a BitBlt.  Of
// course, drawing with transparency is a
// lot more complicated, and the drawing
// code for the control can't use this routine.
//
DWORD CDib::DrawTo( HDC hdc)
{
  BOOL bRet = BitBlt( hdc,
                      0,
                      0,
                      m_lWidth,
                      m_lHeight,
                      m_memDC,
                      0,
                      0,
                      SRCCOPY );

  if ( bRet )
    return 0;
  return GetLastError();
}

