// Simple analysis of profiler information.
// Uses jx.Profiler interface to grab profile information.
//
// This keeps a running count of the number of profiler hits on each line of
// Java code. It also (seperately) keeps track of the number of profiler hits
// in each of the VM internal procedures. Native methods are not counted
// individually, but as a single counter.
//
// When the profile analyzer is running locally, it is (mostly) excluded from
// the analysis, because the getData call (called at the beginning of the update
// procedure) stops the profiler, and it is not restarted until the end of the
// update procedure. Therefore, it is possible to get a few PC's which land in
// the ProfileAnalyzer thread. However, I have never seen it happen. If we
// wanted to, we could ignore points which match the methods in this class.
//
// The data is stored as a hashtable with the method name as the key and
// vectors indexed by line numbers as the value.
// 
// Perhaps a better option is to use the PC's directly as indices into a
// vector-like thing which grows and shrinks dynamically according to the high
// and low bounds. (Why hasn't this been written yet?) Or use a sparse array
// representation.
//
// 19 Jun 1997 John Whaley
//

package jx.profiler;
import java.io.IOException;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;

public class ProfileAnalyzer implements Runnable {

	public int totalPoints;	// total number of profiler points
	public int nativeCnt;	// number of points which are in native code
	public int internalCnt;	// number of points which are in VM internals
	public int javaCnt;		// number of points which are in Java code
	
	public int period;		// number of milliseconds between getData calls

	protected Hashtable jcounters;	// key = Java method name
									// value = table with line #'s and counts

	protected Hashtable icounters;	// key = VM internal procedure name
									// value = count

	protected Profiler prof;		// profiler object

	public boolean stopNotice;		// because jx doesn't support interrupts yet

	// constructor
	public ProfileAnalyzer(Profiler prof)
	{
		javaCnt=internalCnt=nativeCnt=totalPoints=0;
		period=100;
		jcounters = new Hashtable();
		icounters = new Hashtable();
		this.prof = prof;
	}

	// methods to directly control the profiler. (to be used seperately from the thread.)
	public void startProfiling()	{ try {prof.startProfiling();}	catch (IOException e) {}}
	public void stopProfiling()		{ try {prof.stopProfiling();}	catch (IOException e) {}}
	public int size()				{ try {return prof.size();	}	catch (IOException e) {return 0;}}
	public void reset()				{ try {prof.reset();		}	catch (IOException e) {}}

	// thread which updates profile data every once in a while
	public void run()
	{
		try {
			stopNotice = false;
			// open profiler and start analyzing.
			if (!prof.open()) return;
			prof.ignoreNative();
			prof.ignoreVM();
			prof.startProfiling();

			/*** interrupts are not implemented yet in jx ... :(
			for (;;) {
			try {Thread.sleep(period);}	// in milliseconds
			catch (InterruptedException e) {	// interrupt == exit
			update();
			return;
			}
			update();
			startProfiling();
			}
			***/

			for (;;) {
				try {Thread.sleep(period);}	// in milliseconds
				catch (InterruptedException e) {}
				System.out.println(prof.size() + " ticks");
				update();
				if (stopNotice) {
					prof.close();
					stopNotice = false;
					return;	// suicide
				}
				prof.startProfiling();
			}
		} catch (IOException e) {
			System.err.println("Profiler: lost connection");
			prof.close();
			return;
		} // simply exit on IOException
	}

	// update profile data (stops profiler)
	// this is synchronized, but you probably never want to have multiple
	// ProfileAnalyzer threads running at the same time ...
	public synchronized void update_old() throws IOException
	{
		// get the data
		int[] data = prof.getPCData();

		// update counters
		int i, pc;
		ProfilePoint pp;
		totalPoints += data.length;
		for (i=0; i<data.length; ++i) {
			pc = data[i];
			pp = prof.decodePC(pc);
			if (pp.isNative()) {
				// native code, right now do nothing but increment a counter.
				++nativeCnt;
			} else if (pp.isVM()) {
				++internalCnt;
				Integer c = (Integer) icounters.get(pp.methodName);
				int newvalue;
				if (c!=null) newvalue = c.intValue() + 1;
				else newvalue = 1;
				icounters.put(pp.methodName, new Integer(newvalue));
			} else if (pp.isJavaMethod()) {
				++javaCnt;
				// an actual method, now we look it up in the hash table.
				Hashtable v = (Hashtable) jcounters.get(pp.methodName);
				Integer lineNum = new Integer(pp.lineNum);
				if (v!=null) {	// found it
					Integer oldvalue = (Integer)v.get(lineNum);
					int newvalue;
					if (oldvalue != null) newvalue = oldvalue.intValue() + 1;
					else newvalue = 1;
					v.put(lineNum, new Integer(newvalue));
				} else {		// need to add it
					v = new Hashtable (1);
					v.put(lineNum, new Integer(1));
					jcounters.put(pp.methodName, v);
				}
			}
		}
	}

	// update profile data (stops profiler)
	// this is synchronized, but you probably never want to have multiple
	// ProfileAnalyzer threads running at the same time ...
	public synchronized void update() throws IOException
	{
		// get the data
		ProfilePoint[] data = prof.getData();

		// update counters
		int i, pc;
		ProfilePoint pp;
		totalPoints += data.length;
		for (i=0; i<data.length; ++i) {
			pp = data[i];
			if (pp.isNative()) {
				// native code, right now do nothing but increment a counter.
				++nativeCnt;
			} else if (pp.isVM()) {
				++internalCnt;
				Integer c = (Integer) icounters.get(pp.methodName);
				int newvalue;
				if (c!=null) newvalue = c.intValue() + 1;
				else newvalue = 1;
				icounters.put(pp.methodName, new Integer(newvalue));
			} else if (pp.isJavaMethod()) {
				++javaCnt;
				// an actual method, now we look it up in the hash table.
				Hashtable v = (Hashtable) jcounters.get(pp.methodName);
				Integer lineNum = new Integer(pp.lineNum);
				if (v!=null) {	// found it
					Integer oldvalue = (Integer)v.get(lineNum);
					int newvalue;
					if (oldvalue != null) newvalue = oldvalue.intValue() + 1;
					else newvalue = 1;
					v.put(lineNum, new Integer(newvalue));
				} else {		// need to add it
					v = new Hashtable (1);
					v.put(lineNum, new Integer(1));
					jcounters.put(pp.methodName, v);
				}
			}
		}
	}

	// enumerate through methods
	public /*synchronized?*/ Enumeration internalMethods()
	{
		return new ProfileAnalyzerInternalEnumerator(icounters);
	}

	// enumerate through methods
	public /*synchronized?*/ Enumeration javaMethods()
	{
		return jcounters.keys();
	}

	// enumerate through keys (given a method)
	public /*synchronized?*/ Enumeration lines(String methodName)
	{
		Hashtable v = (Hashtable) jcounters.get(methodName);
		if (v != null) return new ProfileAnalyzerJavaEnumerator(v);
		else return new ProfileAnalyzerJavaEnumerator(new Hashtable());
	}

}

class ProfileAnalyzerJavaEnumerator implements Enumeration {

	Enumeration _e;
	Hashtable _v;

	ProfileAnalyzerJavaEnumerator(Hashtable v)
	{
		_v = v;
		_e = v.keys();
	}

	public boolean hasMoreElements()
	{
		return _e.hasMoreElements();
	}

	public Object nextElement()
	{
		Object obj;
		if (!_e.hasMoreElements()) return null;
		obj=_e.nextElement();
		return new Pair(((Integer)obj).intValue(),
			(((Integer)_v.get(obj)).intValue()));
	}

}

class ProfileAnalyzerInternalEnumerator implements Enumeration {

	Enumeration _e;
	Hashtable _v;

	ProfileAnalyzerInternalEnumerator(Hashtable v)
	{
		_v = v;
		_e = v.keys();
	}

	public boolean hasMoreElements()
	{
		return _e.hasMoreElements();
	}

	public Object nextElement()
	{
		Object obj;
		if (!_e.hasMoreElements()) return null;
		obj=_e.nextElement();
		return new InternalPair(((String)obj),
			(((Integer)_v.get(obj)).intValue()));
	}

}
